/******************************************************************************/
/* SlunkCrypt, by LoRd_MuldeR <MuldeR2@GMX.de>                                */
/* This work has been released under the CC0 1.0 Universal license!           */
/******************************************************************************/

/* Internal */
#include "thread.h"
#include "compiler.h"

/* CRT */
#include <stdlib.h>
#include <string.h>

/* PThread */
#if defined(_MSC_VER) && !defined(_DLL)
#  define PTW32_STATIC_LIB 1
#endif
#include <pthread.h>

/* System info */
#ifdef __unix__
#  include <sys/sysinfo.h>
#endif

/* States */
#define TSTATE_IDLE 0U
#define TSTATE_WORK 1U
#define TSTATE_EXIT 2U

// ==========================================================================
// Data types
// ==========================================================================

typedef struct
{
	thrdpl_worker_t worker;
	void *context;
	size_t length;
	uint8_t *buffer;
}
thrdpl_task_t;

typedef struct
{
	size_t thread_count, pending;
	pthread_mutex_t mutex;
	pthread_cond_t cond_pending;
}
thrdpl_shared_t;

typedef struct
{
	thrdpl_shared_t *shared;
	size_t state;
	pthread_cond_t cond_state;
	pthread_t thread;
	thrdpl_task_t task;
}
thrdpl_thread_t;

struct thrdpl_data_t
{
	thrdpl_shared_t shared;
	thrdpl_thread_t thread_data[MAX_THREADS];
};

// ==========================================================================
// Utilities
// ==========================================================================

static INLINE size_t bound(const size_t min, const size_t value, const size_t max)
{
	return (value < min) ? min : ((value > max) ? max : value);
}

#define PTHRD_MUTEX_LOCK(X) do \
{ \
	if (pthread_mutex_lock((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_MUTEX_UNLOCK(X) do \
{ \
	if (pthread_mutex_unlock((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_COND_SIGNAL(X) do \
{ \
	if (pthread_cond_signal((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_COND_BROADCAST(X) do \
{ \
	if (pthread_cond_broadcast((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_COND_WAIT(X,Y) do \
{ \
	if (pthread_cond_wait((X), (Y)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define CHECK_IF_CANCELLED() do \
{ \
	if (data->state == TSTATE_EXIT) \
	{ \
		PTHRD_MUTEX_UNLOCK(&shared->mutex); \
		return NULL; \
	} \
} \
while(0)

// ==========================================================================
// Thread main
// ==========================================================================

static void *worker_thread_main(void *const arg)
{
	thrdpl_thread_t *const data = (thrdpl_thread_t*) arg;
	thrdpl_shared_t *const shared = (thrdpl_shared_t*) data->shared;
	
	thrdpl_task_t *task;

	for (;;)
	{
		PTHRD_MUTEX_LOCK(&shared->mutex);
		CHECK_IF_CANCELLED();

		while (data->state != TSTATE_WORK)
		{
			PTHRD_COND_WAIT(&data->cond_state, &shared->mutex);
			CHECK_IF_CANCELLED();
		}

		task = &data->task;
		PTHRD_MUTEX_UNLOCK(&shared->mutex);

		task->worker(shared->thread_count, task->context, task->buffer, task->length);

		PTHRD_MUTEX_LOCK(&shared->mutex);
		CHECK_IF_CANCELLED();

		data->state = TSTATE_IDLE;
		if (!(--shared->pending))
		{
			PTHRD_COND_BROADCAST(&shared->cond_pending);
		}

		PTHRD_MUTEX_UNLOCK(&shared->mutex);
		PTHRD_COND_SIGNAL(&data->cond_state);
	}
}

// ==========================================================================
// System info
// ==========================================================================

#if defined(__unix__)
#  define NUM_PROCESSORS_FUNC get_nprocs
#elif defined(PTW32_VERSION)
#  define NUM_PROCESSORS_FUNC pthread_num_processors_np
#endif

static size_t detect_cpu_count(void)
{
#ifdef NUM_PROCESSORS_FUNC
	const int cpu_count = NUM_PROCESSORS_FUNC();
	if (cpu_count > 0)
	{
		return (size_t) cpu_count;
	}
#endif
	return 1U;
}

// ==========================================================================
// Manage threads
// ==========================================================================

static int create_worker(thrdpl_shared_t *const shared, thrdpl_thread_t *const thread_data)
{
	thread_data->state = TSTATE_IDLE;
	thread_data->shared = shared;

	if (pthread_cond_init(&thread_data->cond_state, NULL) != 0)
	{
		return -1;
	}

	if (pthread_create(&thread_data->thread, NULL, worker_thread_main, thread_data) != 0)
	{
		pthread_cond_destroy(&thread_data->cond_state);
		return -1;
	}

	return 0;
}

static int destroy_worker(thrdpl_thread_t *const thread_data)
{
	PTHRD_MUTEX_LOCK(&thread_data->shared->mutex);
	thread_data->state = TSTATE_EXIT;
	PTHRD_MUTEX_UNLOCK(&thread_data->shared->mutex);

	PTHRD_COND_BROADCAST(&thread_data->cond_state);
	pthread_join(thread_data->thread, NULL);
	pthread_cond_destroy(&thread_data->cond_state);

	return 0;
}

// ==========================================================================
// Thread pool API
// ==========================================================================

thrdpl_t *slunkcrypt_thrdpl_create(const size_t count)
{
	size_t i, j;
	thrdpl_t *thrdpl = NULL;

	const size_t cpu_count = bound(1U, (count > 0U) ? count : detect_cpu_count(), MAX_THREADS);
	if (cpu_count < 2U)
	{
		return NULL;
	}

	if (!(thrdpl = (thrdpl_t*)malloc(sizeof(thrdpl_t))))
	{
		return NULL;
	}

	memset(thrdpl, 0, sizeof(thrdpl_t));
	thrdpl->shared.thread_count = cpu_count;

	if (pthread_mutex_init(&thrdpl->shared.mutex, NULL) != 0)
	{
		goto failure;
	}
	
	if (pthread_cond_init(&thrdpl->shared.cond_pending, NULL) != 0)
	{
		pthread_mutex_destroy(&thrdpl->shared.mutex);
		goto failure;
	}

	for (i = 0U; i < cpu_count; ++i)
	{
		if (create_worker(&thrdpl->shared, &thrdpl->thread_data[i]) != 0)
		{
			for (j = 0U; j < i; ++j)
			{
				destroy_worker(&thrdpl->thread_data[j]);
			}
			pthread_cond_destroy(&thrdpl->shared.cond_pending);
			pthread_mutex_destroy(&thrdpl->shared.mutex);
			goto failure;
		}
	}

	return thrdpl;

failure:
	free(thrdpl);
	return NULL;
}

size_t slunkcrypt_thrdpl_count(const thrdpl_t *const thrdpl)
{
	return thrdpl->shared.thread_count;
}

void slunkcrypt_thrdpl_exec(thrdpl_t *const thrdpl, const size_t index, const thrdpl_worker_t worker, void *const context, uint8_t *const buffer, const size_t length)
{
	thrdpl_thread_t *const thread = &thrdpl->thread_data[index];

	PTHRD_MUTEX_LOCK(&thrdpl->shared.mutex);

	while ((thread->state != TSTATE_IDLE) && (thread->state != TSTATE_EXIT))
	{
		PTHRD_COND_WAIT(&thread->cond_state, &thrdpl->shared.mutex);
	}

	if (thread->state == TSTATE_EXIT)
	{
		abort(); /*this is not supposed to happen!*/
	}

	thread->state = TSTATE_WORK;
	thread->task.worker = worker;
	thread->task.context = context;
	thread->task.buffer = buffer;
	thread->task.length = length;

	++thrdpl->shared.pending;

	PTHRD_MUTEX_UNLOCK(&thrdpl->shared.mutex);
	PTHRD_COND_SIGNAL(&thread->cond_state);
}

void slunkcrypt_thrdpl_await(thrdpl_t *const thrdpl)
{
	PTHRD_MUTEX_LOCK(&thrdpl->shared.mutex);

	while (thrdpl->shared.pending)
	{
		PTHRD_COND_WAIT(&thrdpl->shared.cond_pending, &thrdpl->shared.mutex);
	}

	PTHRD_MUTEX_UNLOCK(&thrdpl->shared.mutex);
}

void slunkcrypt_thrdpl_destroy(thrdpl_t *const thrdpl)
{
	size_t i;

	if (thrdpl)
	{
		for (i = 0U; i < thrdpl->shared.thread_count; ++i)
		{
			destroy_worker(&thrdpl->thread_data[i]);
		}
		pthread_cond_destroy(&thrdpl->shared.cond_pending);
		pthread_mutex_destroy(&thrdpl->shared.mutex);
		free(thrdpl);
	}
}
